/*
 * When loaded via LD_PRELOAD, disables PI futex support for applications linked
 * against PulseAudio by re-writing pa_mutex_new()'s second argument to false,
 * even if compiled with HAVE_PTHREAD_PRIO_INHERIT set. This is safe to preload
 * even for applications which do not link against PulseAudio.
 *
 * gcc -fPIC -shared -O2 -ldl -o pa_no_prio_inherit.so pa_no_prio_inherit.c
 *
 * GPLv2
 */

#include <dlfcn.h>    // for dlsym, dlclose, dlopen, RTLD_LAZY, RTLD_LOCAL
#include <stdbool.h>  // for true, false, bool
#include <assert.h>   // for assert
#include <glob.h>     // for glob, globfree, GLOB_BRACE
#include <sys/stat.h> // for stat, S_ISDIR
#include <string.h>   // for strdup, NULL

#if defined(__x86_64__)
#  define GLOBSTR "/usr/lib64/libpulsecore-*.so"
#elif defined(__i386__)
#  define GLOBSTR "/usr/lib{32,}/libpulsecore-*.so"
#else
#  error "Unsupported architecture"
#endif

typedef struct pa_mutex pa;

pa *pa_mutex_new(bool, bool);
static char *get_libname(void);

pa *pa_mutex_new(bool recursive, bool inherit_priority __attribute__((unused)))
{
	static pa *(*orig_mutex)(bool, bool);
	char *library;
	void *handle;

	if (!orig_mutex) {
		library = get_libname();
		assert(library);

		handle = dlopen(library, RTLD_LAZY|RTLD_LOCAL);
		assert(handle);

		orig_mutex = dlsym(handle, "pa_mutex_new");
		assert(orig_mutex && !dlclose(handle));
	}

	return orig_mutex(recursive, false);
}

static char *get_libname(void)
{
	glob_t gb;
	struct stat sb;
	char *library;

	if (glob(GLOBSTR, GLOB_BRACE, NULL, &gb) != 0)
		return NULL;

	library = NULL;
	while (gb.gl_pathc-- > 0) {
		if (stat(gb.gl_pathv[gb.gl_pathc], &sb) != 0)
			break;

		if (!S_ISDIR(sb.st_mode)) {
			library = strdup(gb.gl_pathv[gb.gl_pathc]);
			break;
		}
	}

	globfree(&gb);
	return library;
}
